#include "defs.h"
#include <math.h>
#include <stdio.h>
#include "image.h"

#include <cv.h>
#include <highgui.h>

/*
#define MIN(a,b) (a<b) ? a:b
#define MAX(a,b) (a>b) ? a:b
*/

void Tridiagonal3(double** m_aafMat, double* m_afDiag,double* m_afSubd)
{
     double fM00 = m_aafMat[0][0];
     double fM01 = m_aafMat[0][1];
     double fM02 = m_aafMat[0][2];
     double fM11 = m_aafMat[1][1];
     double fM12 = m_aafMat[1][2];
     double fM22 = m_aafMat[2][2];
	 double fQ;
	 double fLength;
     double fInvLength;

     m_afDiag[0] = fM00;
     m_afSubd[2] = 0.0;
     if ( fM02 != 0.0 )
     {
         fLength = sqrt(fM01*fM01+fM02*fM02);
         fInvLength = 1.0/fLength;
         fM01 *= fInvLength;
         fM02 *= fInvLength;
         fQ = 2.0*fM01*fM12+fM02*(fM22-fM11);
         m_afDiag[1] = fM11+fM02*fQ;
         m_afDiag[2] = fM22-fM02*fQ;
         m_afSubd[0] = fLength;
         m_afSubd[1] = fM12-fM01*fQ;
         m_aafMat[0][0] = 1.0; m_aafMat[0][1] = 0.0;  m_aafMat[0][2] = 0.0;
         m_aafMat[1][0] = 0.0; m_aafMat[1][1] = fM01; m_aafMat[1][2] = fM02;
         m_aafMat[2][0] = 0.0; m_aafMat[2][1] = fM02; m_aafMat[2][2] = -fM01;
     }
     else
     {
         m_afDiag[1] = fM11;
         m_afDiag[2] = fM22;
         m_afSubd[0] = fM01;
         m_afSubd[1] = fM12;
         m_aafMat[0][0] = 1.0; m_aafMat[0][1] = 0.0; m_aafMat[0][2] = 0.0;
         m_aafMat[1][0] = 0.0; m_aafMat[1][1] = 1.0; m_aafMat[1][2] = 0.0;
         m_aafMat[2][0] = 0.0; m_aafMat[2][1] = 0.0; m_aafMat[2][2] = 1.0;
     }
}

int QLAlgorithm(int iSize, double* m_afDiag, double* m_afSubd,double** m_aafMat)
{
	int i0,i1,i2,i3,i4;
	double fG,fR,fSin,fCos,fB,fF,fP;
	const int iMaxIter = 32;

     for ( i0 = 0; i0 < iSize; i0++)
     {
         for (i1 = 0; i1 < iMaxIter; i1++)
         {
             for (i2 = i0; i2 <= iSize-2; i2++)
             {
                 double fTmp =
                      fabs(m_afDiag[i2])+  fabs(m_afDiag[i2+1]);
                 if (  fabs(m_afSubd[i2]) + fTmp == fTmp )
                     break;
             }
             if ( i2 == i0 )
                 break;
 
             fG = (m_afDiag[i0+1]-m_afDiag[i0])/(2.0*m_afSubd[i0]);
             fR =  sqrt(fG*fG+1.0);
             if ( fG < 0.0 )
                 fG = m_afDiag[i2]-m_afDiag[i0]+m_afSubd[i0]/(fG-fR);
             else
                 fG = m_afDiag[i2]-m_afDiag[i0]+m_afSubd[i0]/(fG+fR);
             fSin = 1.0;
			 fCos = 1.0;
			 fP = 0.0;
             for (i3 = i2-1; i3 >= i0; i3--)
             {
                 fF = fSin*m_afSubd[i3];
                 fB = fCos*m_afSubd[i3];
                 if (  fabs(fF) >=  fabs(fG) )
                 {
                     fCos = fG/fF;
                     fR = sqrt(fCos*fCos+1.0);
                     m_afSubd[i3+1] = fF*fR;
                     fSin = 1.0/fR;
                     fCos *= fSin;
                 }
                 else
                 {
                     fSin = fF/fG;
                     fR =  sqrt(fSin*fSin+1.0);
                     m_afSubd[i3+1] = fG*fR;
                     fCos = 1.0/fR;
                     fSin *= fCos;
                 }
                 fG = m_afDiag[i3+1]-fP;
                 fR = (m_afDiag[i3]-fG)*fSin+2.0*fB*fCos;
                 fP = fSin*fR;
                 m_afDiag[i3+1] = fG+fP;
                 fG = fCos*fR-fB;
 
                 for (i4 = 0; i4 < iSize; i4++)
                 {
                     fF = m_aafMat[i4][i3+1];
                     m_aafMat[i4][i3+1] = fSin*m_aafMat[i4][i3]+fCos*fF;
                     m_aafMat[i4][i3] = fCos*m_aafMat[i4][i3]-fSin*fF;
                 }
             }
             m_afDiag[i0] -= fP;
             m_afSubd[i0] = fG;
             m_afSubd[i2] = 0.0;
         }
         if ( i1 == iMaxIter )
             return -1;
     }
 
     return 0;
}
void IncreasingSort(int iSize, double* afEigval,double** aafEigvec)
{
	int i0,i1,i2;
	double fMin;
     // sort eigenvalues in increasing order, e[0] <= ... <= e[iSize-1]
     for (i0 = 0, i1; i0 <= iSize-2; i0++)
     {
         // locate minimum eigenvalue
         i1 = i0;
         fMin = afEigval[i1];
         for (i2 = i0+1; i2 < iSize; i2++)
         {
             if ( afEigval[i2] < fMin )
             {
                 i1 = i2;
                 fMin = afEigval[i1];
             }
         }
 
         if ( i1 != i0 )
         {
             // swap eigenvalues
             afEigval[i1] = afEigval[i0];
             afEigval[i0] = fMin;
 
             // swap eigenvectors
             for (i2 = 0; i2 < iSize; i2++)
             {
                 double fTmp = aafEigvec[i2][i0];
                 aafEigvec[i2][i0] = aafEigvec[i2][i1];
                 aafEigvec[i2][i1] = fTmp;
             }
         }
     }
}

int Mean(int **tab,double *mean,int width,int height)
{
	int x,y;
	(*mean) = 0;
	for(y=0 ; y < height ; y++)
	{
		for(x=0 ; x < width ; x++)
		{
			*mean+= tab[y][x];
		}
	}
	*mean /= height * width;
	return 0;
}

double **Covariance(int **tabR,int **tabG,int **tabB,int width,int height)
{
	int x,y;
	int R,G,B;
	double **MatCov;
	double tmpR,tmpG,tmpB,tmpRG,tmpRB,tmpGB;
	double meanR,meanB,meanG;
	MatCov = malloc(sizeof(double*)*3);
	for(x=0 ; x<3 ; x++)
		MatCov[x] = malloc(sizeof(double)*3);

	/*Moyenne des couches*/
	Mean(tabR,&meanR,width,height);
	Mean(tabG,&meanG,width,height);
	Mean(tabB,&meanB,width,height);
	
	tmpR = tmpG = tmpB = tmpRG = tmpRB = tmpGB = 0;
	for(y=0 ; y<height ; y++)
		for(x=0 ; x<width; x++)
		{
			R = tabR[y][x];
			G = tabG[y][x];
			B = tabB[y][x];
			tmpR += (R-meanR)*(R-meanR);
			tmpRG += (R-meanR)*(G-meanG);
			tmpRB += (R-meanR)*(B-meanB);
			tmpG += (G-meanG)*(G-meanG);
			tmpGB += (G-meanG)*(B-meanB);
			tmpB += (B-meanB)*(B-meanB);
		}
	tmpB /= width*height;
	tmpR /= width*height;
	tmpG /= width*height;
	tmpRB /=width*height;
	tmpRG /= width*height;
	tmpGB /= width*height;

	MatCov[0][0] = tmpR;
	MatCov[0][1] = MatCov[1][0] = tmpRG;
	MatCov[0][2] = MatCov[2][0] = tmpRB;
	MatCov[1][1] = tmpG;
	MatCov[2][2] = tmpB;
	MatCov[1][2] = MatCov[2][1] = tmpGB;	
	return MatCov;
}

int MinVar(double **Cov)
{
	double min = MIN(	MIN(Cov[0][0],Cov[1][1]),
				MIN(Cov[1][1],Cov[2][2]));
	if(min == Cov[0][0])
		return 0;
	if(min == Cov[1][1])
		return 1;
	if(min == Cov[2][2])
		return 2;
	return -1;
}
int MaxVar(double **Cov)
{
	double max = MAX(       MAX(Cov[0][0],Cov[1][1]),
	                        MAX(Cov[1][1],Cov[2][2]));
	if(max == Cov[0][0])
		return 0;
	if(max == Cov[1][1])
	        return 1;
	if(max == Cov[2][2])
	        return 2;
	return -1;
}

int acp(image_t *src,image_t *dst)
{
	int x,y;
	double **MatCov;
	double tab1[3],tab2[3];
	int R,G,B,_R,_G,_B;
	int **tabR,**tabG,**tabB;
	tabR = malloc(sizeof(int*)*src->height);
	tabG = malloc(sizeof(int*)*src->height);
	tabB = malloc(sizeof(int*)*src->height);
	/*Construction des plans de couleurs*/
	for(y=0 ; y<src->height ; y++)
	{
		tabR[y] = malloc(sizeof(int)*src->width);
		tabG[y] = malloc(sizeof(int)*src->width);
		tabB[y] = malloc(sizeof(int)*src->width);
		for(x=0 ; x<src->width ; x++)
		{
			getpixel(src,x,y,&R,&G,&B);		
			tabR[y][x] = R;
			tabG[y][x] = G;
			tabB[y][x] = B;
		}
	}
	/*Calcul de la matrice de covariance*/
	MatCov = Covariance(tabR,tabG,tabB,src->width,src->height);

	/*Determination des valeurs et vecteurs propres*/
		/* Simplification de la matrice */
//	printf("%g %g %g\n",MatCov[0][0],MatCov[0][1],MatCov[0][2]);
//	printf("%g %g %g\n",MatCov[1][0],MatCov[1][1],MatCov[1][2]);
//	printf("%g %g %g\n",MatCov[2][0],MatCov[2][1],MatCov[2][2]);
//	printf("\n");
	Tridiagonal3(MatCov,(double*)tab1,(double*)tab2);
//	printf("%g %g %g\n",MatCov[0][0],MatCov[0][1],MatCov[0][2]);
//	printf("%g %g %g\n",MatCov[1][0],MatCov[1][1],MatCov[1][2]);
//	printf("%g %g %g\n",MatCov[2][0],MatCov[2][1],MatCov[2][2]);
		/* Recherche de vecteurs et valeurs propres */
    QLAlgorithm(3,(double*)tab1,(double*)tab2,MatCov);
		/* Classement des vecteurs propres */
    IncreasingSort(3,(double*)tab1,MatCov);
//        printf("\n");
//        printf("%g %g %g\n",MatCov[0][0],MatCov[0][1],MatCov[0][2]);
//	printf("%g %g %g\n",MatCov[1][0],MatCov[1][1],MatCov[1][2]);
//	printf("%g %g %g\n",MatCov[2][0],MatCov[2][1],MatCov[2][2]);

	/*Application de la matrice de passage*/
	for(y=0 ; y< src->height ; y++)
	{
		for(x=0 ; x< src->width; x++)
		{
			getpixel(src,x,y,&R,&G,&B);
			tabR[y][x] = R*MatCov[0][0] + B*MatCov[0][1] + G*MatCov[0][2];
			tabG[y][x] = R*MatCov[1][0] + B*MatCov[1][1] + G*MatCov[1][2];
			tabB[y][x] = R*MatCov[2][0] + B*MatCov[2][1] + G*MatCov[2][2];
		}
	}
	/*Verification de la matrice de covariance*/
	MatCov = Covariance(tabR,tabG,tabB,src->width,src->height);
	/*Recherche de la plus faible variance*/
	_R=_G=_B = 0;
	switch(MinVar(MatCov))
	{
		case 0:
			_R = 1;
			_G = 0;
			_B = 0;
			break;
		case 1:
			_R = 0;
			_G = 1;
			_B = 0;
			break;
		case 2:
			_R = 0;
			_G = 0;
			_B = 1;
			break;
		default:
			printf("Erreur de variance\n");
			break;
	}
	/*Ecriture de l'image resultante*/
	for(y=0 ; y< src->height ; y++)
	{
		for(x=0 ; x< src->width; x++)
		{
			getpixel(src,x,y,&R,&G,&B);
			setpixel(dst,x,y,0,0,R*_R + G*_G + B*_B);
		}
	}
	return 0;
}
int ng(image_t *src,image_t *dst)
{
	int x,y;
	int R,G,B;
	/*Moyenne des plans*/
	for(y=0 ; y<src->height ; y++)
	{
		for(x=0 ; x<src->width ; x++)
		{
			getpixel(src,x,y,&R,&G,&B);		
			setpixel(dst,x,y,0,0,(R+G+B)/3);
		}
	}	
	return 0;
}
