#ifdef _WIN32
typedef unsigned char uint8_t;
#else
#include <stdint.h>
#endif

#include <cv.h>
#include <stdio.h>

#include "defs.h"
#include "image.h"
#include "main.h"

#include <math.h>

#include <cv.h>
#include <highgui.h>

#include "geometry.h"


#define TRANS_X 41
#define TRANS_Y 1 

#define LIMIT_X ((TRANS_X)/2+1)

#define CORR_X 3 
#define CORR_Y 3

#define WINTRANS_X (CORR_X+TRANS_X-1)
#define WINTRANS_Y (CORR_Y+TRANS_Y-1)

/* Continuity weight (16 == normal) */
#define X_WEIGHT 16
#define Y_WEIGHT 16

typedef struct _window Window;
struct _window
{
    int r,g,b;
};

static pt2d_t inverse_transform(projection_t *data,
                                pt2d_t from)
{
    pt2d_t to;
    pt3d_t xx, yy, cm;

    xx = multiply(from.x, data->vx);
    yy = multiply(from.y, data->vy);

    cm = add(data->Os, add(xx, yy));

    to.x = cm.x * data->d / cm.z;
    to.y = cm.y * data->d / cm.z;

    return to;
}

static int getpixel_fast_antialiasing(image_t *img,projection_t*data,
		               double x,double y,int offsetx,int offsety,
			       int *r,int *g,int *b)
{
    double i0,j0,i1,j1,I_,J_;
    int pR,pB,pG;
    double c1,c2,c3,c4;
    pt2d_t from,to;

    from.x = x;
    from.y = y;
    /* XXX: antialiasing */            
    to = inverse_transform(data, from);

   
    /*Pourcentage de couleur */
    j0 = to.x;
    i0 = to.y;
    j1 = to.x + 1.0;
    i1 = to.y + 1.0;

    J_ = (int)to.x;
    I_ = (int)to.y;

    c1 = ((I_+1) - i0) * ((J_+1) - j0);
    c2 = ((I_+1) - i0) * (j1 - (J_+1));
    c3 = (i1 - (I_+1)) * (j1 - (J_+1));
    c4 = (i1 - (I_+1)) * ((J_+1) - j0);
    pR = pB = pG = 0;		
    getpixel_fast(img,(int)to.x+offsetx,(int)to.y+offsety,r,g,b);
    pR = c1 * (*r);
    pG = c1 * (*g);
    pB = c1 * (*b);		
    getpixel_fast(img,(int)to.x+1+offsetx,(int)to.y+offsety,r,g,b);
    pR += c2 * (*r);
    pG += c2 * (*g);
    pB += c2 * (*b);		
    getpixel_fast(img,(int)to.x+1+offsetx,(int)to.y+1+offsety,r,g,b);
    pR += c3*(*r);
    pG += c3*(*g);
    pB += c3*(*b);
    getpixel_fast(img,(int)to.x+offsetx,(int)to.y+1+offsety,r,g,b);
    pR += c4*(*r);
    pG += c4*(*g);
    pB += c4*(*b);		

    *r = pR;
    *g = pG;
    *b = pB;
    
    return 0;
}
static int Calcul(Window *Refer,Window *Trans,int u,int v)
{
    int i,j;
    int corr = 0;
    int delta;
    Window P1,P2; 
    for(j = -CORR_Y/2 ; j < CORR_Y-CORR_Y/2; j++)
    {
        int k = j*CORR_X;
        int l = (j+v)*WINTRANS_X;
        for(i = -CORR_X/2 ; i < CORR_X-CORR_X/2; i++)
        {
            P1 = Refer[k + i]; 
            P2 = Trans[l+i+u]; 
            /*Distance euclidienne*/
            delta = sqrt((P1.r-P2.r)*(P1.r-P2.r)+
			 (P1.g-P2.g)*(P1.g-P2.g)+
			 (P1.b-P2.b)*(P1.b-P2.b));
            corr += delta;
//fprintf(stderr, "TA GUEULE %i (%i %i)\n", delta, i, j);
        }
    }
//fprintf(stderr, "HAHAHAHAHA TA GUEULE %i\n", corr);
    return corr;
}

int Translation_vole(image_t *left , projection_t *left_norm_data, 
	             image_t *right, projection_t *right_norm_data,
                     image_t *left_depth, image_t *right_depth,
                     int p, int q, double adjust_y,
        	     Window *Refer,Window *Trans,int borneinf)
{
    int u,v;
    pt2d_t pt2d;
    int corr,corr_min=0x1000000;
    int umin;
    int r,g,b;
    int depth;
    int uold_x,uold_y;
    int good_x,good_y;
    /* Premiere colonne */
    if(p == 0)
        uold_x = 0;
    else
    {
        getpixel(left_depth, p - 1, q, &r, &g, &b);
        uold_x = r - LIMIT_X;
        if(r != g || uold_x < - LIMIT_X / 2 || uold_x > LIMIT_X - LIMIT_X / 2)
            uold_x = 0;
        else
            good_x = 1;
    }
    /* Premiere ligne */
    if(q == 0)
        uold_y = 0;
    else
    {
        getpixel(left_depth, p, q - 1, &r, &g, &b);
        uold_y = r - LIMIT_X;
        if(r != g || uold_y < - LIMIT_X / 2 || uold_y > LIMIT_X - LIMIT_X / 2)
            uold_y = 0;
        else
            good_y = 1;
    }
    /* Remplissage de la fenetre de translation */
     for(v = -WINTRANS_Y/2; v < WINTRANS_Y-WINTRANS_Y/2; v++)
    {
        int k = v*WINTRANS_X;
    	for(u = -WINTRANS_X/2; u < WINTRANS_X-WINTRANS_X/2; u++)
        {
   	    /*Point normé*/
            pt2d.x = u + p+(int)right_norm_data->left;
    	    pt2d.y = v + q+(int)right_norm_data->top;
            /*dénormalisation d'un point*/
    	    getpixel_fast_antialiasing(right,right_norm_data,
	    		               pt2d.x,pt2d.y,0,0,&r,&g,&b);
            (Trans+k+u)->r = r;
            (Trans+k+u)->g = g;
            (Trans+k+u)->b = b;

	}
    }
    /* Parcours de la translation de gauche a droite */
    for(v = -TRANS_Y/2; v < TRANS_Y-TRANS_Y/2; v++)
    {
        for(u = -TRANS_X/2; u < TRANS_X-TRANS_X/2; u++)
        {
            corr = 16*Calcul(Refer, Trans,u,v);
       	    if(good_x)
            {
                corr += X_WEIGHT * (abs(u - uold_x*0));
                if(good_y)
                    corr += Y_WEIGHT * (abs(u - uold_y+uold_x));
            }
	    /*Minisation de la corrélation*/
            if(corr < corr_min)
            {
		corr_min = corr;
                umin = u;
            }
 	}
    }
    depth = umin + LIMIT_X;
    if(corr_min >= 4000)
        setpixel(left_depth, p, q, 255, 0, 0);
    else
        setpixel(left_depth, p, q, depth, depth, depth);
    return corr_min;
}
int Correlation_vole( image_t *left , projection_t *left_norm_data,
		      image_t *right, projection_t *right_norm_data,
		      image_t *left_depthmap_norm ,
		      image_t *right_depthmap_norm, double adjust)
{
    int p,q;
    int r,g,b;
    pt2d_t pt2d;
    double min,max;
    int pixel; 
    int i,j;
    int Tab_Stat[20],progress;
    int corr;
    Window *Refer;
    Window *Trans;
   
    /*Allocation des fenetres de parcours*/
    Refer = (Window*)malloc(sizeof(Window)*CORR_Y*CORR_X);
    Trans = (Window*)malloc(sizeof(Window)*WINTRANS_Y*WINTRANS_X);
    /*Changement d'origine des pointeurs*/
    Refer = Refer+CORR_X*(CORR_Y/2)+CORR_X/2; 
    Trans = Trans+WINTRANS_X*(WINTRANS_Y/2)+WINTRANS_X/2; 
    /*Initialisation des images resultat*/
    cvFillImage( left_depthmap_norm->_private, CV_RGB(0,0,255));
    cvFillImage( right_depthmap_norm->_private, CV_RGB(0,0,255));
    /* Initialisation du tableau des statistiques */
    for(p = 0; p < 20; p++)
        Tab_Stat[p] = 0;
    for(progress=0,q = 0; q < left_norm_data->bottom- left_norm_data->top ; q++)
    {
        while( q* 100 / (left_norm_data->bottom- left_norm_data->top) >= progress)
        {
            printf("\r %i %% effectués", progress);
            fflush(stdout);
            progress ++;
        }
        for(p = 0; p < left_norm_data->right-left_norm_data->left; p++)
        {
  	    /*Remplissage de la fenetre de reference */
            for(i = -CORR_X/2; i < CORR_X-CORR_X/2; i++)
            {
                for(j = -CORR_Y/2; j < CORR_Y-CORR_Y/2; j++)
	        {
 	            /*Point normé*/
		    pt2d.x = i + p+(int)left_norm_data->left;
	            pt2d.y = j + q+(int)left_norm_data->top;
	    	    /*Dénormalisation d'un point*/
	           /*Récupération couleur dans l'image d'origine à gauche*/
		   getpixel_fast_antialiasing(left , left_norm_data,
				              pt2d.x,pt2d.y,0,0,
					      &r,&g,&b);
		   (Refer+j*CORR_X+i)->r = r;
		   (Refer+j*CORR_X+i)->g = g;
		   (Refer+j*CORR_X+i)->b = b;
		}
            }
	    corr = Translation_vole(left , left_norm_data,
                             right, right_norm_data,
			     left_depthmap_norm, right_depthmap_norm,
                             p, q, 0,Refer,Trans,0);
            Tab_Stat[corr / 1500]++;
	}
    }
    printf("100 %% effectués\n");
    /*Statistiques de la correlation*/
    printf("Statistiques de correlation par pixel\n");
    for(p=0 ; p<20;p++)
        printf("%d - %d => %d\n",p * 1500, (p+1) * 1500, Tab_Stat[p]);
    /*Normalisation de l'affichage*/
    for(min = 999999, max = -999999, p = 0; p <left_depthmap_norm->width ; p++)
    {
        for(q = 0; q < left_depthmap_norm->height ; q++)
        {
            getpixel(left_depthmap_norm, p, q, &r, &g, &b);
            if(r == g)
            {
                if(r < min)
                    min = r;
                if(/*r < 20 &&*/ r > max)
                    max = r;
            }
        }
    }
    printf("min/max: %g %g\n",min,max);
    for(p = 0; p < left_depthmap_norm->width ; p++)
    {
        for(q = 0; q < left_depthmap_norm->height; q++)
        {
            getpixel(left_depthmap_norm, p, q, &r, &g, &b);
            if(r == g)
            {
                pixel = ((float)(max-r)/(float)(max-min))*255.;
                setpixel(left_depthmap_norm, p, q, pixel, pixel, pixel);
            }

            getpixel(right_depthmap_norm, p, q, &r, &g, &b);
            if(r == g)
            {
                pixel = ((float)(max-r)/(float)(max-min))*255.;
                setpixel(right_depthmap_norm, p, q, pixel, pixel, pixel);
            }
        }
    }
    /*Sauvegarde des images resultats*/
    cvSaveImage("sortie.bmp", left_depthmap_norm->_private);
    return 0;
}
