/*
 *  stereo -- Stereo matching software
 *  $Id: delaunaycv.c 167 2004-07-12 08:33:44Z eric $
 *  Copyright (c) 2004 Sam Hocevar <sam@zoy.org>
 *                     Éric Beets <ericbeets@free.fr>
 *  All Rights Reserved
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA
 */

#include <stdint.h>

#include <cv.h>
#include <stdio.h>

#include "defs.h"
#include "image.h"
#include "main.h"

#include <math.h>

#include <cv.h>
#include <highgui.h>

#include "geometry.h"

static pt2d_t transform(projection_t *data,pt2d_t from);                        
double Profondeur(double D,double delta,double x_l,double x_r);
int Delaunay(image_t * depthmap,projection_t *pointnorm,double *tabdepth);



int Carte_3D(	projection_t *left_data,
				projection_t *right_data,
				image_t *left_depthmap,image_t *right_depthmap)
{
	int i;
	double depth;
	double delta;
	double min,max;
	double *tabdepth;
	int tabdepth_count;
	ptsing_t pt2d;

	/* Allocation */
	projection_t *left_pointnorm = malloc(sizeof(projection_t));
	projection_t *right_pointnorm = malloc(sizeof(projection_t));
	/*Initialisation*/	
	left_pointnorm->ptsing = NULL;
	right_pointnorm->ptsing = NULL;
	delta = 10;
	tabdepth = NULL;
	tabdepth_count = 0;	
	cvFillImage( left_depthmap->_private, CV_RGB(255,255,255));
	cvFillImage( right_depthmap->_private, CV_RGB(255,255,255));
	/*Normalisation des points (transformation)*/
	for(i=0 ; i<left_data->ptsing_count ; i++)
	{
		if(left_data->ptsing[i].score >=1)
		{
			/*Image de gauche*/
			pt2d.point.x = left_data->ptsing[i].point.x;		
			pt2d.point.y = left_data->ptsing[i].point.y;
			pt2d.point = transform(left_data,pt2d.point);
			pt2d.point.x -= left_data->left;		
			pt2d.point.y -= left_data->top;		

			array_append(left_pointnorm->ptsing,pt2d);
			/*Image de gauche*/
			pt2d.point.x = right_data->ptsing[i].point.x;
			pt2d.point.y = right_data->ptsing[i].point.y;
			pt2d.point = transform(right_data,pt2d.point);
			pt2d.point.x -= right_data->left;		
			pt2d.point.y -= right_data->top;		
 			array_append(right_pointnorm->ptsing,pt2d);
		}
	}
	min = 9999999;
	max = -9999999;	
	/*Calcul de la profondeur de chaque point*/
	for(i=0 ; i<left_pointnorm->ptsing_count ; i++)
	{
		depth = Profondeur(left_data->D,delta,left_pointnorm->ptsing[i].point.x,right_pointnorm->ptsing[i].point.x);	
		array_append(tabdepth,depth);
		if(depth<min)
			min = depth;
		if(depth>max)
			max = depth;				
	}		
	printf("min : %g\n",min);
	printf("max : %g\n",max);
	/*Ajustement et ecriture de la profondeur*/
	for(i=0 ; i<left_pointnorm->ptsing_count ; i++)
	{
		tabdepth[i]  -= min;
		tabdepth[i] *= 4;
		//tabdepth[i]  = ((tabdepth[i]+fabs(min))/(fabs(max)+fabs(min)))*255;
	}
	/*Construction de Delaunay pour l'image de GAUCHE*/
	Delaunay(left_depthmap,left_pointnorm,tabdepth);
	/*Construction de Delaunay pour l'image de DROITE*/
	Delaunay(right_depthmap,right_pointnorm,tabdepth);
	return 0;
}
				

double Profondeur(double D,double delta,double x_l,double x_r)
{
//	return D*(delta/(x1-x2) - 1);
	return (x_r-x_l);
}

static pt2d_t transform(projection_t *data,
                        pt2d_t from)
{
    pt3d_t pt3d;
    pt2d_t to;

    pt3d.x = from.x;
    pt3d.y = from.y;
    pt3d.z = data->d;

    pt3d = multiply(1 / scalar_product(pt3d, data->Os), pt3d);
    pt3d = multiply(scalar_product(data->Os, data->Os), pt3d);

    to.x = scalar_product(pt3d, data->vx);
    to.y = scalar_product(pt3d, data->vy);

    return to;
}

CvMemStorage* storage;
CvSubdiv2D* subdiv;

CvPoint3D64d Normale(CvPoint3D64d a,CvPoint3D64d b,CvPoint3D64d c,double *d)
{
	CvPoint3D64d normal;
	normal.x  = (b.y - a.y) * (c.z - a.z) - (b.z - a.z) * (c.y - a.y);
	normal.y  = (b.z - a.z) * (c.x - a.x) - (b.x - a.x) * (c.z - a.z);
	normal.z  = (b.x - a.x) * (c.y - a.y) - (b.y - a.y) * (c.x - a.x);
	(*d) = -normal.x*a.x - normal.y*a.y - normal.z*a.z;
	return normal;
}

double Z(CvPoint3D64d pt,CvPoint3D64d normal,double d)
{
	double z;
	z = (-d-normal.x*pt.x - normal.y*pt.y)/normal.z;
	return z;
}


CvSubdiv2D* init_delaunay( CvMemStorage* storage, CvRect rect )
{
    CvSubdiv2D* subdiv;
    
    subdiv = cvCreateSubdiv2D( CV_SEQ_KIND_SUBDIV2D, sizeof(*subdiv),
                               sizeof(CvSubdiv2DPoint),
                               sizeof(CvQuadEdge2D),
                               storage );
    cvInitSubdivDelaunay2D( subdiv, rect );
    
    return subdiv;
}

CvPoint2D32f* SearchPoint(CvPoint2D32f pt)
{
	int i;
	CvSubdiv2DEdge e;
	CvSubdiv2DEdge e0 = 0;
	CvSubdiv2DPoint* p = 0;
	CvSubdiv2DPoint* pt_vois;
	CvPoint2D32f *voisin;
	voisin = malloc(sizeof(CvPoint2D32f)*3);
	cvSubdiv2DLocate( subdiv, pt, &e0, &p );
	if( e0 )
	{
		e = e0;
		i=0;
		do
		{	
			e = cvSubdiv2DGetEdge(e,CV_NEXT_AROUND_LEFT);
			pt_vois = cvSubdiv2DEdgeOrg( e);
		        if( pt_vois )
		        {
			        voisin[i].x = pt_vois->pt.x;
			        voisin[i].y = pt_vois->pt.y;
			        i++;
			}
		}
		while( e != e0 );
	}		
	return voisin;
}


void AddPointDelaunay(image_t *img,projection_t *img_data,double *tabdepth)
{
	int i = 0;
	CvPoint2D32f pt;
	int color;
	for(i=0;i<img_data->ptsing_count;i++)	
	{
		pt.x = img_data->ptsing[i].point.x;	
		pt.y = img_data->ptsing[i].point.y;	
		if(pt.x < subdiv->topleft.x || pt.x > subdiv->bottomright.x || pt.y<subdiv->topleft.y || pt.y>subdiv->bottomright.y)
			continue;
		/* Ajout du point */
	        cvSubdivDelaunay2DInsert( subdiv, pt );
		color = tabdepth[i];		
		SetPixel(img,(int)pt.x,(int)pt.y,color,color,color);
	}
}

int Delaunay(image_t * depthmap,projection_t *pointnorm,double *tabdepth)
{
	double d;
	double color;
	CvPoint3D64d normal;
	uchar R,G,B;
	int i,j,k;
	CvPoint2D32f* triangle,pt;
	CvPoint3D64d tri[3],p;
	CvRect rect = { 0, 0, depthmap->width, depthmap->height };
	storage = cvCreateMemStorage(0);
	subdiv = init_delaunay( storage, rect );
	/*Ajout des */
	 AddPointDelaunay(depthmap,pointnorm,tabdepth);
	for(i=0;i<depthmap->height ; i++)
	{
		for(j=0;j<depthmap->width ; j++)
		{
			/* Recherche du triangle du point */
			pt.x = j;
			pt.y = i;
			triangle = SearchPoint(pt);
			/*Recherche de la profondeur des voisins */	
			for(k=0;k<3;k++)
			{
				tri[k].x = triangle[k].x;	
				tri[k].y = triangle[k].y;	
				GetPixel(depthmap,(int)tri[k].x, (int)tri[k].y ,&R,&G,&B);
				tri[k].z = R;
			}	
			normal = Normale(tri[0],tri[1],tri[2],&d);
			p.x = pt.x;
			p.y = pt.y;
			p.z = Z(p,normal,d);
			color = p.z;
			SetPixel(depthmap,(int)p.x,(int)p.y,(uchar)color,(uchar)color,(uchar)color);
		}
	}
	return 0;
}
			
