/*
 *  stereo -- Stereo matching software
 *  $Id: geometry.c 160 2004-07-02 12:11:09Z sam $
 *  Copyright (c) 2004 Sam Hocevar <sam@zoy.org>
 *                     Éric Beets <ericbeets@free.fr>
 *  All Rights Reserved
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA
 */

#include <math.h>

#include "defs.h"

double distance(pt2d_t a, pt2d_t b)
{
    double dx = a.x - b.x;
    double dy = a.y - b.y;
    return sqrt(dx * dx + dy * dy);
}

pt2d_t intersection(pt2d_t a, pt2d_t b, pt2d_t c, pt2d_t d)
{
    double a1, a2, b1, b2;
    pt2d_t i;

    i.x = i.y = -1;
    if(a.x == b.x || c.x == d.x)
        return i;

    a1 = (a.y - b.y) / (a.x - b.x);
    a2 = (c.y - d.y) / (c.x - d.x);

    if(a1 == a2)
        return i;

    b1 = b.y - a1 * b.x;
    b2 = d.y - a2 * d.x;
    i.x = (b2 - b1) / (a1 - a2);
    i.y = a2 * i.x + b2;

    return i;
}

pt2d_t barycentre(pt2d_t *points, int count)
{
    int i;
    pt2d_t a;

    a.x = a.y = 0.0;

    for(i = 0; i < count; i++)
    {
        a.x += points[i].x;
        a.y += points[i].y;
    }

    if(count)
    {
        a.x /= count;
        a.y /= count;
    }

    return a;
}

double norm(pt3d_t u)
{
    return sqrt(u.x * u.x + u.y * u.y + u.z * u.z);
}

pt3d_t add(pt3d_t u, pt3d_t v)
{
    pt3d_t w;
    w.x = u.x + v.x;
    w.y = u.y + v.y;
    w.z = u.z + v.z;
    return w;
}

pt3d_t sub(pt3d_t u, pt3d_t v)
{
    pt3d_t w;
    w.x = u.x - v.x;
    w.y = u.y - v.y;
    w.z = u.z - v.z;
    return w;
}

pt3d_t multiply(double l, pt3d_t u)
{
    u.x *= l;
    u.y *= l;
    u.z *= l;
    return u;
}

double scalar_product(pt3d_t u, pt3d_t v)
{
    return u.x * v.x + u.y * v.y + u.z * v.z;
}

pt3d_t vector_product(pt3d_t u, pt3d_t v)
{
    pt3d_t w;
    w.x = u.y * v.z - u.z * v.y;
    w.y = u.z * v.x - u.x * v.z;
    w.z = u.x * v.y - u.y * v.x;
    return w;
}

