/*
 *  stereo -- Stereo matching software
 *  $Id: matching.c 208 2004-08-02 15:18:05Z eric $
 *  Copyright (c) 2004 Sam Hocevar <sam@zoy.org>
 *                     Éric Beets <ericbeets@free.fr>
 *  All Rights Reserved
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA
 */

#include <stdint.h>
#include <stdio.h>

#include "defs.h"
#include "main.h"
#include "image.h"


typedef enum _matching_mode matching_mode;

enum _matching_mode
{
	delaunay,
	corr_normal,
	corr_vole,
	corr_bande
};

matching_mode Matching_mode = corr_bande;

/* Perform stereo matching of an image pair */
int matching(pair_t *pair)
{
    image_new(&pair->depthmap_norm[LEFT], pair->image_norm[LEFT].width,
              pair->image_norm[LEFT].height, 3);
    image_new(&pair->depthmap_norm[RIGHT], pair->image_norm[RIGHT].width,
              pair->image_norm[RIGHT].height, 3);
#if 1
    image_new(&pair->depthmap[LEFT], pair->image[LEFT].width,
              pair->image[LEFT].height, 3);
    image_new(&pair->depthmap[RIGHT], pair->image[RIGHT].width,
              pair->image[RIGHT].height, 3);
#else
    image_new_16(&pair->depthmap[LEFT], pair->image[LEFT].width,
              pair->image[LEFT].height, 3);
    image_new_16(&pair->depthmap[RIGHT], pair->image[RIGHT].width,
              pair->image[RIGHT].height, 3);
#endif

	printf("Construction carte 3D\n");
	switch(Matching_mode)
	{
		case delaunay:
			Carte_3D(	&pair->projection[LEFT],
						&pair->projection[RIGHT],
						&pair->depthmap_norm[LEFT],&pair->depthmap_norm[RIGHT]);
			break;
		case corr_normal:
			Correlation_normal(&pair->image_norm[LEFT],
						&pair->image_norm[RIGHT],
						&pair->depthmap_norm[LEFT],
						&pair->depthmap_norm[RIGHT],pair->adjust);
			break;
		case corr_vole:
			Correlation_vole(	&pair->image[LEFT], &pair->projection[LEFT],
								&pair->image[RIGHT],&pair->projection[RIGHT],
								&pair->depthmap_norm[LEFT],
								&pair->depthmap_norm[RIGHT],pair->adjust);
			break;
		case corr_bande:
			Correlation_bande(	&pair->image[LEFT], &pair->projection[LEFT],
								&pair->image[RIGHT],&pair->projection[RIGHT],
								&pair->depthmap_norm[LEFT],
								&pair->depthmap_norm[RIGHT],pair->adjust);
			break;
	}
    image_display(&pair->image_norm[LEFT]);
    image_display(&pair->image_norm[RIGHT]);
    return 0;
}

