/*
 *  stereo -- Stereo matching software
 *  $Id: singularities.c 227 2004-08-13 09:36:36Z eric $
 *  Copyright (c) 2004 Sam Hocevar <sam@zoy.org>
 *                     Éric Beets <ericbeets@free.fr>
 *  All Rights Reserved
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA
 */

#include <math.h>
#include <stdint.h>

#include "defs.h"
#include "main.h"
#include "image.h"


#include <stdio.h>
#include <cv.h>
#include <highgui.h>

#define WIN_SIZE	9

int track_mode = TRACKV4;
int nb_sing;
double seuil;

int InitTrackerBlock(image_t *img,projection_t *img_data,int *nbpoints,int win_size);
int InitTracker(image_t *img,projection_t *img_data,int *nbpoints);
int Track(image_t *img1,projection_t *img1_data,image_t *img2,projection_t *img2_data);

void ComparaisonV1(projection_t *data1,projection_t *data2);
void ComparaisonV2(projection_t *data1,projection_t *data2,double seuil);
void ComparaisonV3(projection_t *data1,projection_t *data2,double seuil);
void ComparaisonV4(projection_t *dataL0,projection_t *dataL1,projection_t *dataR0,projection_t *dataR1,double seuil);

/* Find singular points in the left and right images */
int singularities(image_t *left, projection_t *left_data, image_t *right, projection_t *right_data)
{
    int i;
    image_t *left_tmp = malloc(sizeof(image_t));
    image_t *right_tmp = malloc(sizeof(image_t));
    
    projection_t *right_tmp_data = malloc(sizeof(projection_t));
    projection_t *   left_tmp_data = malloc(sizeof(projection_t));
    
    image_new(left_tmp,left->width,left->height,left->depth);
    image_new(right_tmp,right->width,right->height,right->depth);
    
    cvCopyImage(left->_private,left_tmp->_private);
    cvCopyImage(right->_private,right_tmp->_private);
    
	seuil = 2*sqrt(2);
	nb_sing = 50;

	printf("Taille image : %dx%d\n",((IplImage*)left->_private)->width,((IplImage*)left->_private)->height);
	
	switch(track_mode)
	{
		case TRACKV1:
			/*Initialisation dans l'image de gauche*/
			InitTrackerBlock(left,left_data,&nb_sing,WIN_SIZE);
			/*Suivi dans l'image de droite*/
			Track(left, left_data, right, right_data);
			/*Tous les points sont selectionnes (sauf ceux perdus ...)*/
			ComparaisonV1(left_data,right_data);	
			break;
		case TRACKV2:
			/*Initialisation dans l'image de gauche*/
			InitTrackerBlock(left,left_data,&nb_sing,WIN_SIZE);
			/*Suivi dans l'image de droite*/
			Track(left, left_data, right, right_data);
			/*Suivi dans l'image de gauche*/
			Track(right, right_data, left_tmp, left_tmp_data);
			/*Comparaison des deux images de gauche*/
			ComparaisonV2(left_data,left_tmp_data,seuil);
			/*Mise à jour de l'image de droite*/
			for(i=0 ; i<left_data->ptsing_count ; i++)
				right_data->ptsing[i].score = left_data->ptsing[i].score;
			break;
		case TRACKV3:
			/*Initialisation image de gauche */
			InitTrackerBlock(left,left_data,&nb_sing,WIN_SIZE);
			/*Suivi dans l'image de droite*/
			Track(left, left_data, right, right_data);
			/*Initisalition dans l'image de droite*/
			InitTrackerBlock(right,right_tmp_data,&nb_sing,WIN_SIZE);
			/*Comparaison des deux images de droite*/
			ComparaisonV3(right_data,right_tmp_data,seuil);
			/*Mise à jour des scores de l'image de gauche*/
			for(i=0;i<right_data->ptsing_count;i++)
				left_data->ptsing[i].score = right_data->ptsing[i].score;			
			break;
		case TRACKV4:
			/*Initialisation dans l'image de gauche*/
			InitTrackerBlock(left,left_data,&nb_sing,WIN_SIZE);
			/*Suivi dans l'image de droite*/
			Track(left, left_data, right, right_data);
			/*Initialisation dans l'image de droite*/
			InitTrackerBlock(right_tmp,right_tmp_data,&nb_sing,WIN_SIZE);
			/*Suivi dans l'image de gauche*/
			Track(right_tmp, right_tmp_data, left_tmp, left_tmp_data);
			/*Comparaison des deux images de gauche*/
			ComparaisonV3(left_data,left_tmp_data,seuil);
			/*Comparaison des deux images de droite*/
			ComparaisonV3(right_data,right_tmp_data,seuil);
			/*Combination des images droites et gauche résultats*/
			ComparaisonV4(left_data,left_tmp_data,right_data,right_tmp_data, seuil);
	}
#if 0
//// REMOVE THAT CODE
//			sprintf(buffer,"%d",i);
//			cvPutText( tmp, buffer, cvPoint((int)GetElemX(tab,i),(int)GetElemY(tab,i)), &font, CV_RGB(0,0,255) );
	{
		char buffer[64];
		int x,y;
		double score;
		int color_text=CV_RGB(255,255,255);
		int color_score_sup=CV_RGB(0,255,0);
		int color_score_inf=CV_RGB(0,0,255);
		CvFont font;
		cvInitFont( &font, CV_FONT_VECTOR0, (double).25, (double).25, (double)0, 1 );
		for(i = 0; i < left_data->ptsing_count; i++)
	    { 
	    	/*Image GAUCHE  0 */
	    	x = left_data->ptsing[i].point.x;
	    	y = left_data->ptsing[i].point.y;
	    	score = left_data->ptsing[i].score;
	    	if(score >= 1)
				cvCircle(left->_private,cvPoint(	x,y ),2,color_score_sup,CV_FILLED);
			else
				cvCircle(left->_private,cvPoint(	x, y),2,color_score_inf,0);
			sprintf(buffer,"%d",i);				
			cvPutText( left->_private, buffer, cvPoint(x,y), &font, color_text);
			/*Image DROITE 0 */
	    	x = right_data->ptsing[i].point.x;
	    	y = right_data->ptsing[i].point.y;
	    	score = right_data->ptsing[i].score;
	    	if(score >= 1)    						
		 cvCircle(right->_private,cvPoint(	x, y),2,color_score_sup,CV_FILLED);
			else
				cvCircle(right->_private,cvPoint(	x,y),2,color_score_inf,0);		
			sprintf(buffer,"%d",i);				
			cvPutText( right->_private, buffer, cvPoint(x,y), &font, color_text );
	    	/*Image GAUCHE  1 */
	    	x = left_tmp_data->ptsing[i].point.x;
	    	y = left_tmp_data->ptsing[i].point.y;
	    	score = left_tmp_data->ptsing[i].score;
	    	if(score >= 1)
				cvCircle(left_tmp->_private,cvPoint(	x,y ),2,color_score_sup,CV_FILLED);
			else
				cvCircle(left_tmp->_private,cvPoint(	x, y),2,color_score_inf,0);
			sprintf(buffer,"%d",i);				
			cvPutText( left_tmp->_private, buffer, cvPoint(x,y), &font, color_text );
		/*Image DROITE 1 */
	    	x = right_tmp_data->ptsing[i].point.x;
	    	y = right_tmp_data->ptsing[i].point.y;
	    	score = right_tmp_data->ptsing[i].score;
	    	if(score >= 1)    																	    
				cvCircle(right_tmp->_private,cvPoint(	x, y),2,color_score_sup,CV_FILLED);
			else
				cvCircle(right_tmp->_private,cvPoint(	x,y),2,color_score_inf,0);		
			sprintf(buffer,"%d",i);				
			cvPutText( right_tmp->_private, buffer, cvPoint(x,y), &font, color_text );

	      }
	}
//	cvNamedWindow("test",0);
//	cvNamedWindow("test1",0);
//	cvShowImage("test",left->_private);
//	cvShowImage("test1",right->_private);
//	cvWaitKey(0);
//	image_display(left);
//	image_display(right);
//	image_display(left_tmp);
//	image_display(right_tmp);
//// ----------------
//
	cvSaveImage("left_singuliers0.png",left->_private);
	cvSaveImage("left_singuliers1.png",left_tmp->_private);
	cvSaveImage("right_singuliers0.png",right->_private);
	cvSaveImage("right_singuliers1.png",right_tmp->_private);
#endif
    return 0;
}

/*Decoupage par bloc*/
int InitTrackerBlock(image_t *img,projection_t *img_data,int *nbpoints,int win_size)
{
	int i,j,k;
	int offsetx,offsety;
	int sub = (int)sqrt(win_size);
	image_t  *tmp;
	projection_t *tmp_data;
	int cpt=0;
	int pttotal = 0;
	tmp = malloc(sizeof(image_t));
	tmp_data = malloc(sizeof(projection_t));
	img_data->ptsing = NULL;
	//printf("fenetres : %d\n",sub);
	for(i=0 ; i<sub ; i++)
		for(j=0 ; j<sub; j++)
		{
			offsetx = (int)(j*img->width/sub);
			offsety = (int)(i*img->height/sub);
			GetSubImage(img,tmp,offsetx,offsety,(int)(img->width/sub),(int)(img->height/sub)); 			
			cpt = *nbpoints;
			InitTracker(tmp,tmp_data,&cpt);
			//printf("tracking (%d%d) => %d\n",i,j,cpt);
			pttotal += cpt;
			/*Ajout des points courants trouves*/
			if(!cpt)
				continue;
			for(k=0 ; k<tmp_data->ptsing_count ; k++)
			{
			    ptsing_t pt;
			    pt.point.x = tmp_data->ptsing[k].point.x + offsetx;
			    pt.point.y = tmp_data->ptsing[k].point.y + offsety;
			    pt.score =  tmp_data->ptsing[k].score;
			    array_append(img_data->ptsing,pt);
			}
		}	
		return 0;
}

/*Determination des points singuliers */
int InitTracker(image_t *img,projection_t *img_data,int *nbpoints)
{
	/*Declaration*/
	int i;
	CvPoint2D32f *tmp;
	IplImage *temp[2];
//	IplImage *img0;
//	image_t *temp[2];
	image_t *img0;
	CvSize size;
	ptsing_t pt;
    
	/*Initialisation*/
	img_data->ptsing = NULL;
	size.width = img->width;
	size.height = img->height;
	
    /*Allocation*/
    tmp = (CvPoint2D32f*)malloc((*nbpoints)*sizeof(CvPoint2D32f));
//    img0 = cvCreateImage(size,IPL_DEPTH_8U,1);
//    temp[0] = malloc(sizeof(image_t));
//    temp[1] = malloc(sizeof(image_t));
    img0 = malloc(sizeof(image_t));
    image_new(img0,size.width,size.height,1);
	for(i=0;i<2;i++)
	{
		temp[i] = cvCreateImage(size,IPL_DEPTH_32F,1);
		temp[i]->depth = IPL_DEPTH_32F;
      temp[i]->nChannels = 1;

//		image_new(temp[i],size.width,size.height,1);
//		((IplImage*)temp[i]->_private)->depth = IPL_DEPTH_32F;
//		((IplImage*)temp[i]->_private)->nChannels = 1;
//		temp[i]->depth = 1;
		
    }    
    /*Conversion NG*/
//   	cvCvtColor(img->_private,img0,CV_BGR2GRAY);
	convert_BGR2GRAY(img,img0);
	/*Détermination des points singuliers */
	//cvGoodFeaturesToTrack((IplImage*)(img0->_private),(IplImage*)(temp[0]->_private),(IplImage*)(temp[1]->_private),tmp,nbpoints,0.1,10,0);
	cvGoodFeaturesToTrack(img0->_private,temp[0],temp[1],tmp,nbpoints,0.1,10,0);
	if( *nbpoints > 0)
	{
		/*Affinement des points singuliers */
		cvFindCornerSubPix( img0->_private, tmp, *nbpoints, cvSize(5,5), cvSize(-1,-1), cvTermCriteria( CV_TERMCRIT_ITER, 10, 0.1f ));
		/* Ajout dans le tableau final (status ==0 <=> point non suivi) */
		for(i=0;i<*nbpoints ; i++)
		{
			pt.point.x = tmp[i].x;
			pt.point.y = tmp[i].y;
			pt.score = 1.0;
			array_append(img_data->ptsing,pt);
		}
	}
		/* Libération de la mémoire */
		//cvReleaseImage(&img0);
		image_free(img0);
		for(i=0;i<2;i++)
		{
			cvReleaseImage(&temp[i]);
//			image_free(temp[i]);
		}
		free(tmp);
	return 0;
}
/*Suivi des points */
int Track(image_t *img1,projection_t *img1_data,image_t *img2,projection_t *img2_data)
{
	int i;
	int pyr0_ready = 0;
	//IplImage *frame1,*frame0,*pyr1,*pyr0;
	IplImage *pyr1,*pyr0;
	image_t *frame1,*frame0;
	char *status;
	CvPoint2D32f *tmp0,*tmp1;
	CvSize size;
	ptsing_t pt;
	int count = img1_data->ptsing_count;
	size.width = img1->width;
	size.height = img1->height;
	/* Allocation et Création des images en NG */
//	frame0 = cvCreateImage(size,IPL_DEPTH_8U,1);
//	frame1 = cvCreateImage(size,IPL_DEPTH_8U,1);
	frame0 = malloc(sizeof(image_t));
	frame1 = malloc(sizeof(image_t));
	image_new(frame0,size.width,size.height,1);
	image_new(frame1,size.width,size.height,1);
	
	pyr0 = cvCreateImage(size,IPL_DEPTH_8U,1);
	pyr1 = cvCreateImage(size,IPL_DEPTH_8U,1);
	
	tmp0 = (CvPoint2D32f*)malloc(count*sizeof(CvPoint2D32f));
	tmp1 = (CvPoint2D32f*)malloc(count*sizeof(CvPoint2D32f));
	status = (char*)malloc(sizeof(char)*count);
	/* Conversion NB */
	//cvCvtColor(img1->_private,frame0,CV_BGR2GRAY);
	//cvCvtColor(img2->_private,frame1,CV_BGR2GRAY);
	convert_BGR2GRAY(img1,frame0);
	convert_BGR2GRAY(img2,frame1);
	/* Récupération des points */
	for(i=0;i<count;i++)
	{
		tmp0[i].x = (float)img1_data->ptsing[i].point.x;
		tmp0[i].y = (float)img1_data->ptsing[i].point.y;
	}
	/* Suivi des points */
	 cvCalcOpticalFlowPyrLK(	frame0->_private, frame1->_private, pyr0, pyr1,
								tmp0,tmp1,
								count, cvSize(10,10), 3,
								status, 0, 
								cvTermCriteria(CV_TERMCRIT_ITER|CV_TERMCRIT_EPS,20,0.03),
								pyr0_ready ? CV_LKFLOW_PYR_A_READY : 0 );
	
	img2_data->ptsing = NULL;
	img2_data->ptsing_count = 0;
	for(i=0 ; i<count ; i++)
	{
		pt.point.x = tmp1[i].x;
		pt.point.y = tmp1[i].y;
		if(status[i] == 0)
		{
			pt.score=-1;
			img1_data->ptsing[i].score = -1;
		}
		else
		{
			img1_data->ptsing[i].score = 1;
			pt.score=1;
		}
		array_append(img2_data->ptsing,pt);
	}			
	/* libération mémoire */
	free(tmp0);
	free(tmp1);
	free(status);
//	cvReleaseImage(&frame0);
//	cvReleaseImage(&frame1);
	image_free(frame0);
	image_free(frame1);
	cvReleaseImage(&pyr0);
	cvReleaseImage(&pyr1);
	printf("fin suivi\n");
	return 0;
}

/*Comparaison*/
double Distance(double x1,double y1,double x2,double y2)
{
	return sqrt( (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2));
}
void ComparaisonV1(projection_t *data1,projection_t *data2)
{	
	int i;	
	for(i=0;i<data1->ptsing_count;i++)
	{
		/*Point perdu*/
		if((data1->ptsing[i].score == -1) || (data2->ptsing[i].score == -1))
		{
			data1->ptsing[i].score = -1;
			data2->ptsing[i].score = -1;	
		}
		else
		{
			data1->ptsing[i].score = 1;
			data2->ptsing[i].score = 1;
		}
	}
}
void ComparaisonV2(projection_t *left0,projection_t *left1,double seuil)
{
	int i;
	double d;
	for(i=0 ; i<left0->ptsing_count ; i++)
	{
		/*Point perdu*/
		if((left0->ptsing[i].score == -1) || (left1->ptsing[i].score == -1))
		{
				left0->ptsing[i].score = -1;
				left1->ptsing[i].score = -1;			
		}		
		if((left0->ptsing[i].score == 1) && (left1->ptsing[i].score == 1))
		{
			d = Distance(left0->ptsing[i].point.x,left0->ptsing[i].point.y,left1->ptsing[i].point.x,left1->ptsing[i].point.y);
			if( d <= seuil)
			{
				/*point suivi*/
				left0->ptsing[i].score = 1;
				left1->ptsing[i].score = 1;
			}
			else
			{
				/*point non suivi*/
				left0->ptsing[i].score = 0;
				left1->ptsing[i].score = 0;			
			}
		}		
	}	
}
/* Comparaison de deux tableau non-ordonnés (images identiques */
void ComparaisonV3(projection_t *right0,projection_t *right1,double seuil)
{
	int i,j;
	int index_min_j;
	double dist_min;
	double d;
	/* Pour la premiere image */
	for(i=0;i<right0->ptsing_count;i++)
	{
		dist_min = 9999999;
		index_min_j = -1;
		/* Pour la seconde image => recherche du point mini distant*/
		for(j=0;j<right1->ptsing_count;j++)
		{
			if(right0->ptsing[i].score == 1 && right1->ptsing[j].score == 1)
			{
				d = Distance(	right0->ptsing[i].point.x,right0->ptsing[i].point.y,
								right1->ptsing[j].point.x,right1->ptsing[j].point.y);
				if(d<seuil && d < dist_min)
				{
					dist_min = d;
					index_min_j = j;
				}
			}
		}
		if(	index_min_j !=right1->ptsing_count && index_min_j !=-1)
		{
			right0->ptsing[i].score = 2;
			right1->ptsing[index_min_j].score = 2;		
		}
	}
	/* Normalisation sur 1*/	
	for(i=0;i<right0->ptsing_count;i++)
	{
		if(right0->ptsing[i].score == 2)
			right0->ptsing[i].score = 1;
		else
			right0->ptsing[i].score = 0;
		// Attention .. memoire...	
		if(right1->ptsing[i].score == 2)
			right1->ptsing[i].score = 1;
		else
			right1->ptsing[i].score = 0;
	}	
}
/* Comparaison des indices */
void ComparaisonV4(projection_t *left0,projection_t *left1,projection_t *right0,projection_t *right1,double seuil)
{
	int i;
	for(i=0 ; i<left0->ptsing_count ; i++)
	{
		if(left0->ptsing[i].score != 1 || right0->ptsing[i].score != 1)
		{
			left0->ptsing[i].score = -1;
			right0->ptsing[i].score = -1;
		}
	}
	for(i=0 ; i<left1->ptsing_count ; i++)
	{	
		if(left1->ptsing[i].score != 1 || right1->ptsing[i].score != 1)
		{
			left1->ptsing[i].score = -1;
			right1->ptsing[i].score = -1;
		}
	}
}
